--This Tank Stops for Nobody! - A scenario inspired by Never Stop the Train
--Written by Mylon, 2018
--MIT License, https://opensource.org/licenses/MIT

ttsfn = {}
ttsfn.fuels = {wood = 0.2, ["small-electric-pole"] = 0.2, ["raw-wood"] = 0.2, coal = 0.22, ["solid-fuel"] = 0.25, ["rocket-fuel"] = 0.26, ["nuclear-fuel"] = 0.28 }
ttsfn.surface = {MARGIN = 64, HEIGHT = 128, START_WIDTH = 96, START_HEIGHT = 64}
ttsfn.COAST = 0.2
ttsfn.STARTING_MAX_HP = 100000
ttsfn.RAW_MAX_HP = 2000
ttsfn.MULTITANK_MODE = true
ttsfn.RESOURCES = {"iron-ore", "copper-ore", "coal", "stone", "crude-oil"}
ttsfn.MINING_POWER = 10
ttsfn.MINING_ENERGY_CONSUMED = 180000 / 3 * 0.9 --Approximate amount of energy to power a drill to produce 1 ore.
--ttsfn.RAW_MAX_HP = game.entity_prototypes["tank"].max_health --Tank prototype hp
--global.ttsfn.players[player.name] = {tank=entity, raw_hp_last_tick=float, hp=float, max_hp=float, pos=position, open_spawn=bool, mining_power=int, internal_size=int, auto_turn=float, windbane=bool}
global.ttsfn = {tanks = {}, players = {}}

function ttsfn.drive(event)
    --Process each tank
    for _, player in pairs(game.connected_players) do
        --Move forward, deduct fuel.
        if not global.ttsfn.players[player.name] then log("Player not initialized.") return end

        --This step will skip over crew members.
        local tank = global.ttsfn.players[player.name].tank
        if not (tank and tank.valid) then 
            global.ttsfn.players[player.name].tank = nil
        else
            
            --Tank externals
            --Hit a cliff or water?
            if tank.speed == 0 then
                local orientation = (0.25 - tank.orientation) * 2 * math.pi
                local dx, dy = math.cos( orientation ) * 0.02, -math.sin( orientation ) * 0.02
                tank.teleport({tank.position.x + dx, tank.position.y + dy})
            end

            --Set speed and deduct fuel
            local speed = 0
            local drivetrain = global.ttsfn.players[player.name].drivetrain
            if drivetrain.energy < drivetrain.electric_drain then
                if tank.burner.currently_burning then
                    local burner = tank.burner
                    speed = ttsfn.fuels[burner.currently_burning.name]
                    if not speed then
                        speed = 0
                        burner.remaining_burning_fuel = 0
                        log("Unrecognized fuel type.")
                    end
                    burner.remaining_burning_fuel = burner.remaining_burning_fuel - 800000 / 60
                    if burner.remaining_burning_fuel < 0 then
                        --if burner.inventory[1] and burner.inventory[1].valid and not burner.inventory[1].is_empty() then --May have to check all fuel inventory slots
                        if burner.inventory[1] and burner.inventory[1].valid_for_read then --May have to check all fuel inventory slots
                            burner.currently_burning = burner.inventory[1].prototype
                            burner.inventory[1].count = burner.inventory[1].count - 1
                            burner.remaining_burning_fuel = burner.inventory[1].prototype.fuel_value
                        else
                            speed = ttsfn.COAST
                        end
                    end
                end
            else
                speed = 0.1 * (0.1 + 0.9 * math.max(1, drivetrain.energy / drivetrain.electric_buffer_size))
            end

            tank.speed = speed

            if global.ttsfn.players[player.name].windvane then
                tank.orientation = tank.orientation + tank.surface.wind_orientation_change
            else
                tank.orientation = tank.orientation + global.ttsfn.players[player.name].auto_turn
            end
            
            -- Modify the hp
            -- TODO: Consume repair packs to increase hp.
            local dhp = tank.health - global.ttsfn.players[player.name].raw_hp_last_tick
            global.ttsfn.players[player.name].hp = global.ttsfn.players[player.name].hp + dhp
            --We can't intercept the death event, so the tank has to stay above one-shot territory.  Nukes can still probably bypass this, but players shouldn't be able to fire them anyway.
            tank.health = math.max(100, global.ttsfn.players[player.name].hp / global.ttsfn.players[player.name].max_hp * ttsfn.RAW_MAX_HP)
            
            global.ttsfn.players[player.name].raw_hp_last_tick = tank.health

            
            --Periodic functions.  We don't need to do this every tick.
            if (game.tick + player.index) % 30 == 0 then
                --Detect and deplete nearby ore.
                local mining_power = global.ttsfn.players[player.name].mining_power
                for k,v in pairs(ttsfn.RESOURCES) do
                    local area = {{tank.position.x-10, tank.position.y-10}, {tank.position.x+10, tank.position.y+10}}
                    local ore = tank.surface.find_entities_filtered{name=v, area=area, limit=1}[1]
                    if ore and ore.valid then
                        --Spawn particles.
                        tank.surface.create_entity{name="electric-beam", duration=30, position=tank.position, target=ore.position, source=tank}

                        --Add ore, deduct mining power from budget
                        --ore to add = mining power * force.mining_drill_productivity_bonus
                        local chest =  global.ttsfn.players[player.name].chests[v]
                        chest.insert{name=v, count=math.floor(mining_power * (1+tank.force.mining_drill_productivity_bonus)) }

                        --Deplete ore.
                        if ore.amount <= mining_power then
                            script.raise_event(defines.events.on_resource_depleted, {entity=ore, name=defines.events.on_resource_depleted})
                            if ore and ore.valid then
                                ore.destroy()
                            end
                        else
                            ore.amount = ore.amount - mining_power
                        end
                    end
                end
            
                --In case no one is driving, request to generate chunks so the biters can get their noms on.
                local chunkx, chunky = math.floor(tank.position.x / 32), math.floor(tank.position.y / 32)
                for x = -10, 10 do
                    for y = -10, 10 do
                        if not tank.surface.is_chunk_generated({chunkx + x, chunky + y}) then
                            if x > -2 and x < 2 and y > -2 and y < 2 then
                                tank.surface.set_chunk_generated_status({chunkx + x, chunky + y}, defines.chunk_generated_status.custom_tiles)
                            else
                                tank.surface.set_chunk_generated_status({chunkx + x, chunky + y}, defines.chunk_generated_status.entities)
                            end
                        end
                    end
                end
                tank.surface.request_to_generate_chunks({chunkx, chunky}, 2)

                --Do upgrades.
                ttsfn.upgrade(player)
                
                --End periodic functions.
            end



            --Tank internals
            global.ttsfn.players[player.name].water_tank.fluidbox[1] = {name="water", amount=25000}
            global.ttsfn.players[player.name].oil_tank.fluidbox[1] = {name="crude-oil", amount=25000}

            --Every second, collect pollution inside the tank and create it outside.

        end

    end
end

--Each player spawns as a tank.
function ttsfn.create_player(event)
    local player = game.players[event.player_index]
    
    --For single-tank mode, second player and up, skip this and move them inside of the tank.
    if not ttsfn.MULTIPLAYER_MODE and event.player_index > 1 then
        player.teleport({-2, ttsfn.surface.START_HEIGHT / 2 + 6}, "ttsfn")
        return
    end

    --Tank
    local tank = game.surfaces[1].create_entity{name="tank", position = {0,0}, force=game.forces.player}
    local direction = math.random()
    local force = game.forces.player
    tank.insert{name="raw-wood", count=100}
    tank.orientation = direction
    tank.set_driver(player)
    table.insert(global.ttsfn.tanks, tank)
    global.ttsfn.players[player.name] = {tank=tank, max_hp=ttsfn.STARTING_MAX_HP, hp=ttsfn.STARTING_MAX_HP, raw_hp_last_tick=2000}

    --Inner Surface
    local surface = game.surfaces.ttsfn
    local start_x = -1
    local start_y = (event.player_index) * (ttsfn.surface.MARGIN + ttsfn.surface.HEIGHT)
    local tiles = {}
    for x = start_x, -ttsfn.surface.START_WIDTH, -1 do
        for y = start_y, start_y + ttsfn.surface.START_HEIGHT - 1 do
            --Exception: top right triangle remains void.
            if y > start_y + ttsfn.surface.START_HEIGHT / 2 + x then
                table.insert(tiles, {name="concrete", position={x, y}, hidden_tile="brick"})
            end
            global.ttsfn.players[player.name].end_y = y
        end
        global.ttsfn.players[player.name].end_x = x
    end
    surface.set_tiles(tiles)

        
    --Create Gui
    ttsfn.driving_gui(player)

    --Water and oil tanks
    local function protected(entity)
        entity.minable = false
        entity.destructible = false
    end
    global.ttsfn.players[player.name].water_tank = surface.create_entity{name="storage-tank", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + 2}}
    protected(global.ttsfn.players[player.name].water_tank)
    
    global.ttsfn.players[player.name].oil_tank = surface.create_entity{name="storage-tank", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT - 3}}
    protected(global.ttsfn.players[player.name].oil_tank)

    local chest, loader, offset
    --Ore chests
    global.ttsfn.players[player.name].chests = {}
    --Stone
    offset = 13
    chest = surface.create_entity{name="logistic-chest-passive-provider", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    loader = surface.create_entity{name="express-loader", type="output", force=game.forces.player, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}, direction=defines.direction.west}
    loader.rotatable = false
    protected(chest)
    protected(loader)
    global.ttsfn.players[player.name].chests.stone = chest
    force.add_chart_tag(surface, {icon={type="item", name="stone"}, text="Stone output chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --Wood
    offset = 5
    chest = surface.create_entity{name="logistic-chest-passive-provider", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    loader = surface.create_entity{name="express-loader", type="output", force=game.forces.player, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}, direction=defines.direction.west}
    loader.rotatable = false
    protected(chest)
    protected(loader)
    global.ttsfn.players[player.name].chests["raw-wood"] = chest
    force.add_chart_tag(surface, {icon={type="item", name="raw-wood"}, text="Wood output chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --Coal
    offset = 7
    chest = surface.create_entity{name="logistic-chest-passive-provider", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    loader = surface.create_entity{name="express-loader", type="output", force=game.forces.player, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}, direction=defines.direction.west}
    loader.rotatable = false
    protected(chest)
    protected(loader)
    global.ttsfn.players[player.name].chests.coal = chest
    force.add_chart_tag(surface, {icon={type="item", name="coal"}, text="Coal output chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --iron-ore
    offset = 9
    chest = surface.create_entity{name="logistic-chest-passive-provider", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    loader = surface.create_entity{name="express-loader", type="output", force=game.forces.player, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}, direction=defines.direction.west}
    loader.rotatable = false
    protected(chest)
    protected(loader)
    global.ttsfn.players[player.name].chests["iron-ore"] = chest
    force.add_chart_tag(surface, {icon={type="item", name="iron-ore"}, text="Iron ore output chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --copper-ore
    offset = 11
    chest = surface.create_entity{name="logistic-chest-passive-provider", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    loader = surface.create_entity{name="express-loader", type="output", force=game.forces.player, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}, direction=defines.direction.west}
    loader.rotatable = false
    protected(chest)
    protected(loader)
    global.ttsfn.players[player.name].chests["copper-ore"] = chest
    force.add_chart_tag(surface, {icon={type="item", name="copper-ore"}, text="Copper ore output chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --Upgrade-chest
    offset = 15
    chest = surface.create_entity{name="logistic-chest-requester", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}    
    protected(chest)
    global.ttsfn.players[player.name].chests.upgrade = chest
    force.add_chart_tag(surface, {icon={type="item", name="tank"}, text="Upgrade chest", target=chest, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --Drivetrain
    offset = 18
    local drivetrain = surface.create_entity{name="electric-energy-interface", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}
    protected(drivetrain)
    drivetrain.operable = false
    drivetrain.power_production = 0
    drivetrain.power_usage = 800000
    global.ttsfn.players[player.name].drivetrain = drivetrain
    force.add_chart_tag(surface, {icon={type="item", name="electric-energy-interface"}, text="Drivetrain", target=drivetrain, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --Mining drill
    offset = 21
    local mining_drill = surface.create_entity{name="electric-energy-interface", force=game.forces.player, position = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}}
    protected(mining_drill)
    mining_drill.operable = false
    mining_drill.power_production = 0
    mining_drill.power_usage = 0
    global.ttsfn.players[player.name].mining_power = ttsfn.MINING_POWER
    drivetrain.electric_buffer_size = global.ttsfn.players[player.name].mining_power * 1000000 --1 MJ per mining power.
    global.ttsfn.players[player.name].mining_drill = mining_drill
    force.add_chart_tag(surface, {icon={type="item", name="electric-energy-interface"}, text="Mining-Drill power", target=mining_drill, position = {start_x-3, start_y + ttsfn.surface.START_HEIGHT / 2 + offset}})

    --This controls how long it takes for low power to affect the tank.
    drivetrain.electric_buffer_size = drivetrain.power_usage * 10
    global.ttsfn.players[player.name].drivetrain = drivetrain

    --Initialize variables.
    global.ttsfn.players[player.name].auto_turn = 0
    global.ttsfn.players[player.name].pos = {start_x-2, start_y + ttsfn.surface.START_HEIGHT / 2 + 6}
    

    --Create map tags for information.

end

function ttsfn.nom(event)
    if not (event.entity and event.entity.valid and event.entity.type == "tree") then return end
    local tank = event.entity.surface.find_entities_filtered{name="tank", limit=1, area={{event.entity.position.x-2, event.entity.position.y-2}, {event.entity.position.x+2, event.entity.position.y+2}} }[1]
    if tank and tank.valid then
        --Find the matching tank and inser into it's wood chest.
        for k,v in pairs(global.ttsfn.players) do
            if v.tank and v.tank.valid and v.tank == tank then
                v.chests["raw-wood"].insert({name="raw-wood", count=4})
                --tank.insert({name="raw-wood", count=4})
            end
        end
    end
end

function ttsfn.upgrade(player)
    --Consume inputs from upgrade chest and improve related stats.

    --Electric drills -> Boost mining power, drivetrain power draw

    --Locomotives -> Boost max speed, drivetrain power draw

    --Cargo wagons -> boost inside size.
    while global.ttsfn.players[player.name].chests.upgrade.get_item_count("cargo-wagon") >= 8 do
        ttsfn.expand(player)
    end

    --Tanks -> Boost turret compartment size

    --Heavy armor -> Boost max HP


end

function ttsfn.expand(player)
    --Deduct 1 cargo wagon per 8 tiles.
    global.ttsfn.players[player.name].chests.upgrade.remove_item{name="cargo-wagon", count=8}

    global.ttsfn.players[player.name].end_x = global.ttsfn.players[player.name].end_x - 1
    local start_y = (player.index) * (ttsfn.surface.MARGIN + ttsfn.surface.HEIGHT)
    local end_x = global.ttsfn.players[player.name].end_x
    local tiles = {}
    for y = start_y, start_y + ttsfn.surface.START_HEIGHT - 1 do
        table.insert(tiles, {name="concrete", hidden_tile="brick", position={end_x, y}})
    end
    game.surfaces.ttsfn.set_tiles(tiles)

end

--Gui create
function ttsfn.driving_gui(player)
    if player.gui.top["ttsfn-driving"] then return end --Somehow this got called twice.
    local pane = player.gui.top.add{type="frame", name="ttsfn-driving", direction="vertical"}
    pane.add{type="button", name="ttsfn-enter", caption="Inside"}
    pane.add{type="button", name="ttsfn-stats", caption="Stats"}
    pane.add{type="checkbox", name="ttsfn-open", caption="Accept Crew", state=false}
    local turn = pane.add{type="flow", direction="horizontal", name="turn"}
    turn.add{type="slider", name="ttsfn-autoturn", orientation="horizontal", caption="Auto-turn", minimum_value=-0.0005, maximum_value=0.0005}
    turn.add{type="checkbox", name="ttsfn-windvane", caption="Random", state=false}
end

function ttsfn.inside_gui(player)
    if player.gui.top["ttsfn-inside"] then return end
    local pane = player.gui.top.add{type="frame", name="ttsfn-inside", direction="vertical"}
    pane.add{type="button", name="ttsfn-drive", caption="Drive"}
    pane.add{type="button", name="ttsfn-stats", caption="Stats"}
end

--Stats!
function ttsfn.stats(player)
    --Display info like speed (current, max)

    --Mining power

    --hp / max_hp

    --size

end

--Gui functions.
function ttsfn.gui_click(event)
    if not (event.element and event.element.valid) then return end
    local player = game.players[event.player_index]
    if event.element.name == "ttsfn-enter" then
        event.element.parent.destroy()
        ttsfn.inside_gui(player)
        player.teleport(global.ttsfn.players[player.name].pos, "ttsfn")
        return
    end

    if event.element.name == "ttsfn-drive" then
        ttsfn.driving_gui(player)
        event.element.parent.destroy()
        global.ttsfn.players[player.name].pos = player.position
        player.teleport({0,0}, "nauvis")
        global.ttsfn.players[player.name].tank.set_driver(player)

        return
    end

end

function ttsfn.auto_turn(event)
    if not (event.element and event.element.valid) then return end
    if not event.element.name == "ttsfn-autoturn" then return end

    local player = game.players[event.player_index]
    global.ttsfn.players[player.name].auto_turn = event.element.slider_value

end

function ttsfn.windvane(event)
    if not (event.element and event.element.valid) then return end
    if not event.element.name == "ttsfn-windvane" then return end
    local player = game.players[event.player_index]
    global.ttsfn.players[player.name].windvane = event.element.state

    event.element.parent["ttsfn-autoturn"].enabled = not event.element.state
end

--Create the TTSFN surface
function ttsfn.on_init()
    --The initial 3x3 area should never be used.  But we can't specify 0 height and width.
    --local surface = game.create_surface("ttsfn", {width=1, height=1})
    local surface = game.create_surface("ttsfn")
    --Suppress normal terrain generation.  Map exclusively uses the third quadrant.
    for x = -50, 20 do
        for y = -20, 50 do
            surface.set_chunk_generated_status({x,y}, defines.chunk_generated_status.entities)
        end
    end

    --Need to chart the surface so map tags can be drawn.
    game.forces.player.chart_all(surface)

    surface.daytime = 0.5
    surface.freeze_daytime = true

    game.forces.enemy.evolution = 0.4
    --For some reason this doesn't work.
    -- for x = -1, 1 do
    --     for y = -1, 1 do
    --         surface.set_tiles{{name="out-of-map", position={x,y}}}
    --     end
    -- end
    --Default permissions does not enable leaving the vehicle!
    local perms = game.permissions.groups[1]
    perms.set_allows_action(defines.input_action.toggle_driving, false)
end

--TODO: Check for low HP to interrupt death and use custom HP
Event.register(defines.events.on_gui_checked_state_changed, ttsfn.windvane)
Event.register(defines.events.on_gui_value_changed, ttsfn.auto_turn)
Event.register(defines.events.on_entity_died, ttsfn.nom)
Event.register(defines.events.on_gui_click, ttsfn.gui_click)
Event.register(defines.events.on_tick, ttsfn.drive)
Event.register(defines.events.on_player_created, ttsfn.create_player)
Event.register(-1, ttsfn.on_init)